/* Nessus Attack Scripting Language
 *
 * Copyright (C) 1999 - 2003 Renaud Deraison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#include <includes.h>
#include <pcap.h>

#include "capture_packet.h"
#include "nasl_raw.h"



extern int islocalhost(struct in_addr *);




/*
 * Set up the pcap filter, and select the correct interface.
 *
 * The filter will be changed only if this is necessary
 * 
 */
 
int init_capture_device(struct in_addr src, struct in_addr dst, char * filter)

{
 int ret;
 char * interface = NULL;
 char * a_dst, *a_src;
 char errbuf[PCAP_ERRBUF_SIZE];
 int free_filter = 0;
 
 a_src = estrdup(inet_ntoa(src));
 a_dst = estrdup(inet_ntoa(dst));
 
 if((filter == NULL) || (filter[0]=='\0') || (filter[0]=='0'))
 {
  filter = emalloc(256);
  free_filter = 1;
  if(islocalhost(&src) == 0)
  	sprintf(filter, "ip and (src host %s and dst host %s)",
 		  a_src, a_dst);
		  
 }
 else {
 	if(islocalhost(&src) == 0)filter = estrdup(filter);
	else filter = emalloc(1);
	free_filter = 1;
 	}
 
 efree(&a_dst);
 efree(&a_src);

 if((interface = routethrough(&src, &dst))||
    (interface = pcap_lookupdev(errbuf)))   
    ret = bpf_open_live(interface, filter);
     
 
 if(free_filter != 0)efree(&filter);
 return ret;   
}



struct ip * capture_next_packet(int bpf, int timeout)
{
  struct pcap_pkthdr head;
  int dl_len;
  char * packet = NULL;
  char * ret = NULL;
  struct timeval past, now;
  struct timezone tz;
 
  if(bpf < 0)
   return NULL;
   
  dl_len =  get_datalink_size(bpf_datalink(bpf));
  bzero(&past, sizeof(past));
  bzero(&now, sizeof(now));
  gettimeofday(&past, &tz);
  for(;;)
  {
   packet = (char*)bpf_next(bpf, &head);
   if(packet != NULL)
     break;
   gettimeofday(&now, &tz);
   if((now.tv_sec - past.tv_sec)>=timeout)break;
  }
  
  if(packet != NULL)
  {
   struct ip * ip;
   ip = (struct ip *)(packet + dl_len);
#ifdef BSD_BYTE_ORDERING
   ip->ip_len = ntohs(ip->ip_len);
   ip->ip_off = ntohs(ip->ip_off);
#endif   
   ip->ip_id = ntohs(ip->ip_id);
   ret = emalloc(UNFIX(ip->ip_len));
   bcopy(ip, ret, UNFIX(ip->ip_len));
  }
 return((struct ip*)ret);
}
