/* Nessus Attack Scripting Language version 2
 *
 * Copyright (C) 2002 Michel Arboi
 *
 * En quelques mots : ceci n'est pas un logiciel libre.
 *
 * Ce logiciel "NASL2" n'est pas termin. Il reste ma proprit jusqu' ce 
 * que je dcide de changer la licence.
 * En attendant, vous avez le droit de linker avec le code sous GPL "Nessus", 
 * provenant de http://www.nessus.org,  fin de tests ou pour votre propre 
 * usage ducatif ou priv.
 * Vous n'avez pas le droit de revendre sous quelque forme que ce soit le
 * logiciel NASL2 ou ses rsultats ; cette restriction s'applique  tout 
 * logiciel link avec NASL2, ou incluant une partie de code provenant de
 * NASL2.
 *
 * Ce programme est distribu en esprant qu'il sera utile, mais SANS AUCUNE
 * GARANTIE.
 * Si vous n'acceptez pas ce point, veuillez me contacter pour acheter une
 * licence commerciale.
 *
 * [quick & dirty english translation. In case of conflict, French version
 *  prevails]
 *
 * In a few words: this is *not* free software.
 *
 * This "NASL2" software is unfinished. It is proprietary software until 
 * I decide to switch to another licence.
 * Meanwhile, you may link it with the GPL-ed "Nessus" software, from 
 * http://www.nessus.org, and use it for testing or for your own edutionnal 
 * or private use.
 * You are not allow to sell either this NASL2 software or its outputs;
 * the same restriction applies to any software linked with NASL2, or that
 * includes a part of NASL2 code.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * If you do not accept this last point, please contact me and buy a commercial
 * licence.
 */
#include <includes.h>

#include "nasl_tree.h"
#include "nasl_global_ctxt.h"
#include "nasl_func.h"
#include "nasl_var.h"
#include "nasl_lex_ctxt.h"
#include "exec.h"

#include "nasl_init.h"
#include "strutils.h"

#ifndef NASL_DEBUG
#define NASL_DEBUG 0
#endif

int
cell2bool(lex_ctxt* lexic, tree_cell* c)
{
  tree_cell	*c2;
  int		flag;
  char		*p;


  if (c == NULL || c == FAKE_CELL)
    return 0;
  
  switch (c->type)
    {
    case CONST_INT:
      return c->x.i_val != 0;
    case CONST_STR:
    case CONST_DATA:
      if (c->size == 0)
	return 0;
      if(c->x.str_val[0] == '0' && c->size == 1)
	  return 0;
     return 1;  

    default:
      c2 = nasl_exec(lexic, c);
      flag = cell2bool(lexic, c2);
      deref_cell(c2);
      return flag;
    }
}

static int
cvt_bool(lex_ctxt* lexic, tree_cell* c)
{
  int	flag;

#if 0
  nasl_dump_tree(c);
#endif
  flag = cell2bool(lexic, c);
  /* free(c); free_tree(c); */
  return flag;
}


int
cell2int(lex_ctxt* lexic, tree_cell* c)
{
  tree_cell	*c2 = NULL;
  int		x;

  if (c == NULL) /*  so that EXPR_MINUS works with unary minus */
    return 0;

  switch(c->type)
    {
    case CONST_INT:
      return c->x.i_val;
    case CONST_STR:
    case CONST_DATA:  
      return strtol(c->x.str_val, NULL, 0);
    default:
      c2 = nasl_exec(lexic, c);
      x = cell2int(lexic, c2);
      deref_cell(c2);
      return x;
    }
}

tree_cell*
int2cell(int x)
{
  tree_cell	*c = alloc_expr_cell(0, CONST_INT, NULL, NULL);
  c->x.i_val = x;
  return c;
}

tree_cell*
bool2cell(int x)
{
  return int2cell(x);
}

char*
cell2str(lex_ctxt* lexic, tree_cell* c)
{
  char	* p;
  tree_cell	*c2;

  if (c == NULL || c == FAKE_CELL)
    {
#if NASL_DEBUG > 0
      fprintf(stderr, "Cannot convert NULL or FAKE cell to string\n");
#endif
      return NULL;
    }
      
  switch(c->type)
    {
    case CONST_INT:
      p = malloc(16);
      sprintf(p, "%d", c->x.i_val);
      return p;
      
    case CONST_STR:
    case CONST_DATA:
      if ( c->x.str_val == NULL)
	p = estrdup("");
      else
	p = strndup(c->x.str_val, c->size);
      return p;
      
    default:
      c2 = nasl_exec(lexic, c);
      p = cell2str(lexic, c2);
      deref_cell(c2);
      if (p == NULL)
	p = estrdup("");
      return p;
    }
}

/* cell2atom returns a 'referenced' cell */	
tree_cell* cell2atom(lex_ctxt* lexic, tree_cell* c1)
{
  tree_cell	*c2 = NULL, *ret = NULL;
  if (c1 == NULL || c1 == FAKE_CELL)
    return c1;

  switch(c1->type)
    {
    case CONST_INT:
    case CONST_STR:
    case CONST_DATA:
      ref_cell(c1);
      return c1;
    default:
      c2 = nasl_exec(lexic, c1);
      ret = cell2atom(lexic, c2);
      deref_cell(c2);
      return ret;
    }
}

int
cell_cmp(lex_ctxt* lexic, tree_cell* c1, tree_cell* c2)
{
  int		flag, x1, x2, typ;
  char		*s1, *s2;
  tree_cell	*d1 = NULL, *d2 = NULL;


#if NASL_DEBUG > 0
  if (c1 == NULL)
    fprintf(stderr, "cell_cmp: c1 == NULL !\n");
  if (c2 == NULL)
    fprintf(stderr, "cell_cmp: c2 == NULL !\n");
#endif

  if (c1 == NULL && c2 == NULL)
     return 0;
     
  if (c1 == NULL || c2 == NULL)
    return -1;			/* does not really matter, IMHO */

  typ = c1->type;
  if((c2->type == CONST_DATA || c2->type == CONST_STR) &&
     (typ != CONST_DATA && typ != CONST_STR)) typ = c2->type;
    
     
  switch (typ)
    {
    case CONST_INT:
      x1 = cell2int(lexic, c1);
      x2 = cell2int(lexic, c2);
      return x1 - x2;

    case CONST_STR:
    case CONST_DATA:
      {
      int len_s1, len_s2;
  
      s1 = cell2str(lexic, c1);
      if(c1->type == CONST_STR ||
         c1->type == CONST_DATA)
	   len_s1 = c1->size;
      else
      	   len_s1 = strlen(s1);
	   
	   	   
      s2 = cell2str(lexic, c2);
      if(c2->type == CONST_STR ||
         c2->type == CONST_DATA)
	   len_s2 = c2->size;
      else
      	   len_s2 = strlen(s2);

		  
      if(len_s1 != len_s2)
       len_s1 - len_s2;
      else	   	   
       flag = memcmp(s1, s2, len_s1);
       	
      efree(&s1); efree(&s2); 
      return flag;
     }
    default:
      d1 = cell2atom(lexic, c1);
      d2 = cell2atom(lexic, c2);
      flag = cell_cmp(lexic, d1, d2);
      deref_cell(d1);
      deref_cell(d2);
      return flag;
    }
}

FILE*	nasl_trace_fp = NULL;

lex_ctxt* truc = NULL;

static void
nasl_dump_expr(FILE* fp, const tree_cell* c)
{
  if (c == NULL)
    fprintf(fp, "NULL");
  else if  (c == FAKE_CELL)
    fprintf(fp, "FAKE");
  else
    switch(c->type)
      {
      case NODE_VAR:
	fprintf(fp, "%s", c->x.str_val);
	break;
      case EXPR_AND:
	fprintf(fp, "(");
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, ") && (");
	nasl_dump_expr(fp, c->link[1]);
	fprintf(fp, ")");
	break;
      case EXPR_OR:
	fprintf(fp, "(");
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, ") || (");
	nasl_dump_expr(fp, c->link[1]);
	fprintf(fp, ")");
	break;
      case EXPR_NOT:
	fprintf(fp, "! (");
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, ")");
	break;
      case EXPR_PLUS:
	fprintf(fp, "(");
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, ") + (");
	nasl_dump_expr(fp, c->link[1]);
	fprintf(fp, ")");
	break;
      case EXPR_MINUS:
	fprintf(fp, "(");
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, ") - (");
	nasl_dump_expr(fp, c->link[1]);
	fprintf(fp, ")");
	break;
      case EXPR_MULT:
	fprintf(fp, "(");
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, ") * (");
	nasl_dump_expr(fp, c->link[1]);
	fprintf(fp, ")");
	break;
      case EXPR_DIV:
	fprintf(fp, "(");
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, ") / (");
	nasl_dump_expr(fp, c->link[1]);
	fprintf(fp, ")");
	break;
      case EXPR_MODULO:
	fprintf(fp, "(");
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, ") % (");
	nasl_dump_expr(fp, c->link[1]);
	fprintf(fp, ")");
	break;
      case EXPR_BIT_AND:
	fprintf(fp, "(");
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, ") & (");
	nasl_dump_expr(fp, c->link[1]);
	fprintf(fp, ")");
	break;
      case EXPR_BIT_OR:
	fprintf(fp, "(");
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, ") | (");
	nasl_dump_expr(fp, c->link[1]);
	fprintf(fp, ")");
	break;
      case COMP_MATCH:
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, " >< ");
	nasl_dump_expr(fp, c->link[1]);
	break;
      case COMP_LT:
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, " < ");
	nasl_dump_expr(fp, c->link[1]);
	break;
      case COMP_LE:
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, " <= ");
	nasl_dump_expr(fp, c->link[1]);
	break;
      case COMP_GT:
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, " > ");
	nasl_dump_expr(fp, c->link[1]);
	break;
      case COMP_GE:
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, " >= ");
	nasl_dump_expr(fp, c->link[1]);
	break;
      case COMP_EQ:
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, " == ");
	nasl_dump_expr(fp, c->link[1]);
	break;
      case CONST_INT:
	fprintf(fp, "%d", c->x.i_val);
	break;
      case CONST_STR:
      case CONST_DATA:
	fprintf(fp, "\"%d\"", c->x.str_val);
	break;

      case NODE_ARRAY_EL:
	fprintf(fp, "%s[", c->x.str_val);
	nasl_dump_expr(fp, c->link[0]);
	fprintf(fp, "]");
	break; 

      case NODE_FUN_CALL:
	fprintf(fp, "%s(...)", c->x.str_val);
	break;
	
      case NODE_AFF:
	nasl_dump_expr(fp, c->link[0]);
	putc('=', fp);
	nasl_dump_expr(fp, c->link[1]);
	break;

      default:
	fprintf(fp, "*%d*", c->type);
	break;      
      }
}

static void
nasl_short_dump(FILE* fp, const tree_cell* c)
{
  if (c == NULL || c == FAKE_CELL)
    return;

  switch (c->type)
    {
    case NODE_IF_ELSE:
      fprintf(fp, "NASL:%04d> if (", c->line_nb);
      nasl_dump_expr(fp, c->link[0]);
      fprintf(fp, ") { ... }");
      if (c->link[2] != NULL)  fprintf(fp, " else { ... }");
      putc('\n', fp);
      break;

    case NODE_FOR:
      fprintf(fp, "NASL:%04d> for (", c->line_nb); nasl_dump_expr(fp, c->link[0]);
      fprintf(fp, "; "); nasl_dump_expr(fp, c->link[1]);
      fprintf(fp, "; "); nasl_dump_expr(fp, c->link[2]);
      fprintf(fp, ") { ... }\n");
      break;

    case NODE_WHILE:
      fprintf(fp, "NASL:%04d> while (", c->line_nb);
      nasl_dump_expr(fp, c->link[0]);
      fprintf(fp, ") { ... }\n");
      break;

    case NODE_REPEAT:
      fprintf(fp, "NASL:%04d> ... x ", c->line_nb); 
      nasl_dump_expr(fp, c->link[1]);
      putc('\n', fp);
      break;

    case NODE_RETURN:
      fprintf(fp, "NASL:%04d> return ", c->line_nb);
      nasl_dump_expr(fp, c->link[0]);
      fprintf(fp, ";\n");
      break;

    case NODE_AFF:
      fprintf(fp, "NASL:%04d> ", c->line_nb);
      nasl_dump_expr(fp, c);
      fprintf(fp, ";\n");
      break;

    case NODE_FUN_CALL:
      fprintf(fp, "NASL:%04d> %s(...)\n", c->line_nb, c->x.str_val);
      break;

    case NODE_LOCAL:
      fprintf(fp, "NASL:%04d> local_var ...\n", c->line_nb);
      break;

    case NODE_GLOBAL:
      fprintf(fp, "NASL:%04d> global_var ...\n", c->line_nb);
      break;
    }
}



tree_cell*
nasl_exec(lex_ctxt* lexic, tree_cell* st)
{
  tree_cell	*ret = NULL, *ret2 = NULL, *tc1 = NULL, *tc2 = NULL, *idx = NULL, *args;
  int		flag, x, y;
  char		*s1 = NULL, *s2 = NULL, *s3 = NULL, *p = NULL;
  nasl_func	*pf = NULL;
  int		i, n, sz;

#if 0
  nasl_dump_tree(st);      /* See rt.value, rt.type, rt.length */
#endif

  if (lexic->ret_val != NULL)
    {
      ref_cell(lexic->ret_val);
      return lexic->ret_val;
    }

  if (st == FAKE_CELL)
    return FAKE_CELL;

  if (st == NULL)
    {
#if NASL_DEBUG > 0
      fprintf(stderr, "nasl_exec: st == NULL\n");
#endif
      return NULL;
    }

  if (nasl_trace_fp != NULL)
    nasl_short_dump(nasl_trace_fp, st);

  switch(st->type)
    {
    case NODE_IF_ELSE:
      if ((ret = nasl_exec(lexic, st->link[0])) == NULL)
	return NULL;
      if (cvt_bool(lexic, ret))
	ret2 = nasl_exec(lexic, st->link[1]);
      else
	if (st->link[2] != NULL) /* else branch */
	  ret2 = nasl_exec(lexic, st->link[2]);
	else			/* No else */
	  ret2 = FAKE_CELL;
      deref_cell(ret);
      return ret2;

    case NODE_INSTR_L:	/* Block. [0] = first instr, [1] = tail */
      ret = nasl_exec(lexic, st->link[0]);
#if NASL_DEBUG > 1
      if (ret == NULL)
	fprintf(stderr, "Instruction failed. Going on in block\n");
#endif
      if (st->link[1] == NULL)
	return ret;
      deref_cell(ret);
      ret = nasl_exec(lexic, st->link[1]);
      return ret;
	
    case NODE_FOR:
      /* [0] = start expr, [1] = cond, [2] = end_expr, [3] = block */
      if ((ret2 = nasl_exec(lexic, st->link[0])) == NULL)
	return NULL;
      deref_cell(ret2);
      for (;;)
	{
	  /* Break the loop if 'return' */
	  if (lexic->ret_val != NULL)
	    {
	      ref_cell(lexic->ret_val);
	      return lexic->ret_val;
	    }
	    
	  /* condition */
	  if ((ret = nasl_exec(lexic, st->link[1])) == NULL)
	    return NULL;
	  flag = cvt_bool(lexic, ret);
	  deref_cell(ret);
	  if (! flag)
	    break;
	  /* block */
	  if ((ret = nasl_exec(lexic, st->link[3])) == NULL)
	    return NULL;
	  deref_cell(ret);
	  /* end expression */
	  if ((ret = nasl_exec(lexic, st->link[2])) == NULL)
	    return NULL;
	  deref_cell(ret); 
	}
      return FAKE_CELL;

    case NODE_WHILE:
      /* [0] = cond, [1] = block */
      for (;;)
	{
	  /* return? */
	  if (lexic->ret_val != NULL)
	    {
	      ref_cell(lexic->ret_val);
	      return lexic->ret_val;
	    }
	  /* Condition */
	  if ((ret = nasl_exec(lexic, st->link[0])) == NULL)
	    return NULL;
	  flag = cvt_bool(lexic, ret);
	  deref_cell(ret);
	  if (! flag)
	    break;
	  if ((ret = nasl_exec(lexic, st->link[1])) == NULL)
	    return NULL;
	  deref_cell(ret);
	}
      return FAKE_CELL;

    case NODE_FUN_DEF:
      /* x.str_val = function name, [0] = argdecl, [1] = block */
      ret = decl_nasl_func(lexic, st);
      return ret;

    case NODE_FUN_CALL:
      pf = get_func_ref_by_name(lexic, st->x.str_val);
      if (pf == NULL)
	{
	  fprintf(stderr, "Undefined function '%s'\n", st->x.str_val);
	  return NULL;
	}
      args = st->link[0];
#if 0
      printf("****************\n");
      nasl_dump_tree(args);
      printf("****************\n");
#endif
      ret = nasl_func_call(lexic, pf, args);
      return ret;

    case NODE_REPEAT:
      n = cell2int(lexic, st->link[1]);
      if (n <= 0)
	return NULL;
	
#ifdef STOP_AT_FIRST_ERROR	
      for (tc1 = NULL, i = 1; i <= n; i ++)
	{
	  deref_cell(tc1);
	  if ((tc1 = nasl_exec(lexic, st->link[0])) == NULL)
	    return NULL;
	}
      return tc1;
#else
      for (i = 1; i <= n; i ++)
	{
	  tc1 = nasl_exec(lexic, st->link[0]);
	  deref_cell(tc1);
	}
      return FAKE_CELL;
#endif

      /*
       * I wonder... 
       * Will nasl_exec be really called with NODE_EXEC or NODE_ARG?
       */
    case NODE_DECL:		/* Used in function declarations */
      /* [0] = next arg in list */
      /* TBD? */
      return st;		/* ? */

    case NODE_ARG:		/* Used function calls */
      /* val = name can be NULL, [0] = val, [1] = next arg */
      ret = nasl_exec(lexic, st->link[0]);	/* Is this wise? */
      return ret;

    case NODE_RETURN:
      /* [0] = ret val */
      ret = nasl_return(lexic, st->link[0]);
      return ret;

    case NODE_ARRAY_EL:		/* val = array name, [0] = index */
      idx = cell2atom(lexic, st->link[0]);
      ret = get_array_elem(lexic, st->x.str_val, idx);
      deref_cell(idx);
      return ret;

    case NODE_AFF:
      /* [0] = lvalue, [1] = rvalue */
      tc1 = nasl_exec(lexic, st->link[0]);
      tc2 = nasl_exec(lexic, st->link[1]);
      ret = nasl_affect(tc1, tc2);
      deref_cell(tc1);		/* Must free VAR_REF */
      deref_cell(ret);
      return tc2;		/* So that "a = b = e;" works */

    case NODE_VAR:
      /* val = variable name */
      ret = get_variable_by_name(lexic, st->x.str_val);
      return ret;

    case NODE_LOCAL:		/* [0] = argdecl */
      ret = decl_local_variables(lexic, st->link[0]);
      return ret;

    case NODE_GLOBAL:		/* [0] = argdecl */
      ret = decl_global_variables(lexic, st->link[0]);
      return ret;

    case EXPR_AND:
      x = cell2bool(lexic, st->link[0]);
      y = cell2bool(lexic, st->link[1]);
      return bool2cell(x && y);

    case EXPR_OR:
      x = cell2bool(lexic, st->link[0]);
      y = cell2bool(lexic, st->link[1]);
      return bool2cell(x || y);

    case EXPR_NOT:
      x = cell2bool(lexic, st->link[0]);
      return bool2cell(! x);

    case EXPR_INCR:
    case EXPR_DECR:
      x =  (st->type == EXPR_INCR) ? 1 : -1;
      if (st->link[0] == NULL)
	{
	  y = 1;		/* pre */
	  tc1 = st->link[1];
	}
      else
	{
	  y = 0;		/* post */
	  tc1 = st->link[0];
	}
      tc2 = nasl_exec(lexic, tc1);
      if (tc2 == NULL)
	return NULL;
      ret = nasl_incr_variable(lexic, tc2, y, x);
      deref_cell(tc2);
      return ret;

      if (st->link[0] == NULL)
	ret = nasl_incr_variable(lexic, st->link[1], 1, 1);
      else
	ret = nasl_incr_variable(lexic, st->link[1], 0, 1);
      break;

    case EXPR_PLUS:
      s1 = s2 = NULL;
      tc1 = cell2atom(lexic, st->link[0]);
      if (tc1 == NULL || tc1 == FAKE_CELL)
	return NULL;

      tc2 = cell2atom(lexic, st->link[1]);
      if (tc2 == NULL || tc1 == FAKE_CELL)
	{
	  deref_cell(tc1);
	  return NULL;
	}
      	
      flag = tc1->type;
      
      if((tc1->type != CONST_STR && tc1->type != CONST_DATA) ||
         (tc2->type != CONST_STR && tc2->type != CONST_DATA ))
      { 
       if(tc2->type == CONST_DATA || tc2->type == CONST_STR)
        {
       	flag = tc2->type;
	s1 = cell2str(lexic, tc1);
	}
      }
      
      switch (flag)
	{
	case CONST_INT:
	  x = tc1->x.i_val;
	  y = cell2int(lexic, st->link[1]);
	  ret = int2cell(x + y);
	  break;
	case CONST_STR:
	case CONST_DATA:
	  if (tc2->type != CONST_STR && tc2->type != CONST_DATA)
	    {
	      s2 = cell2str(lexic, tc2);
	      if (s2 == NULL)
		{
		  ret = NULL;
		}
	      else
		{
		  sz = tc1->size + strlen(s2);
		  s3 = emalloc(sz);
		  bcopy(tc1->x.str_val, s3, tc1->size);
		  bcopy(s2, s3+tc1->size, strlen(s2));
		  efree(&s2);
		  ret = alloc_tree_cell(0, s3);
		  ret->type = flag;
		  ret->size = sz;
		}
	    }
	  else
	    {
	      if( s1 != NULL )
	      {
	       int len = strlen(s1);
	     
	       sz = len + tc2->size;
	       s3 = emalloc(sz);
	       bcopy(s1, s3, len);
	       bcopy(tc2->x.str_val, s3 + len, tc2->size);
	       efree(&s1);
	       ret = alloc_tree_cell(0, s3);
	       ret->type = flag;
	       ret->size = sz;
	     }
	     else
	     {  
	      sz = tc1->size + tc2->size;
	      s3 = emalloc(sz+1);
	      memcpy(s3, tc1->x.str_val, tc1->size);
	      memcpy(s3 + tc1->size, tc2->x.str_val, tc2->size);
	      ret = alloc_tree_cell(0, s3);
	      ret->size = tc1->size + tc2->size;
#if 1
	      if (tc1->type == CONST_DATA || tc2->type == CONST_DATA)
		ret->type = CONST_DATA;
	      else
		ret->type = CONST_STR;
#else
	      ret->type = CONST_DATA;
#endif
	     }
	    }
	  break;

	default:
	  ret = NULL;
	  break;
	}
      deref_cell(tc1);
      deref_cell(tc2);
      return ret;

    case EXPR_MINUS:		/* Infamous duplicated code */
      s1 = s2 = NULL;
      tc1 = cell2atom(lexic, st->link[0]);
      if (tc1 == NULL || tc1 == FAKE_CELL)
	return NULL;
      tc2 = cell2atom(lexic, st->link[1]);
      if (tc2 == NULL || tc1 == FAKE_CELL)
	{
	  deref_cell(tc1);
	  return NULL;
	}
		
      flag = tc1->type;
      
     if((tc1->type != CONST_STR && tc1->type != CONST_DATA) ||
        (tc2->type != CONST_STR && tc2->type != CONST_DATA ))
      {
       if(tc2->type == CONST_STR || tc2->type == CONST_STR)
       {
        s1 = cell2str(lexic, tc1);
	flag = tc2->type;
       }
      }
      
      switch (flag)
	{
	case CONST_INT:
	  x = tc1->x.i_val;
	  y = cell2int(lexic, st->link[1]);
	  ret = int2cell(x - y);
	  break;

	case CONST_STR:
	case CONST_DATA:
	  {
	  int sz_1, sz_2;

	  if (tc2->type != CONST_STR && tc2->type != CONST_DATA)
	    {
	    s2 = cell2str(lexic, tc2);
	    sz_2 = strlen(s2);
	    }
	  else
	   {
	    s2 = tc2->x.str_val;
	    sz_2 = tc2->size;
	   }
	  if (s2 == NULL)
	    {
	      ret = NULL;
	    }
	  else
	    {
	      if(s1 == NULL)
	      {
	       s1 = tc1->x.str_val;
	       sz_1 = tc1->size;
	      }
	      else
	      { 
	       sz_1 = strlen(s1);
	      }
	      
	      p = memmem(s1, sz_1,  s2, sz_2);
	      if (s2[0] == '\0' || p == NULL)
		{
		  s3 = emalloc(sz_1);
		  bcopy(s1, s3, sz_1);
		  ret = alloc_tree_cell(0, s3);
		  ret->type = flag;
		  ret->size = sz_1;
		}
	      else
		{
		  sz = sz_1 - sz_2;
		  if (sz <= 0)
		    {
		      sz = 0; s3 = estrdup("");
		    }
		  else
		    {
		      s3 = emalloc(sz);
		      if (p - s1 > 0)
			memcpy(s3, s1, p - s1);
		      if (sz > p - s1)
			memcpy(s3 + (p - s1), p + sz_2, sz - (p - s1));
		    }
		  ret = alloc_tree_cell(0, s3);
		  ret->size = sz;
		  ret->type = flag;
		}
	      if (s2 != tc2->x.str_val)
		efree(&s2);
	      if (s1 != tc1->x.str_val)
	        efree(&s1);
	    }
	  break;
	 }

	default:
	  ret = NULL;
	  break;
	}
      deref_cell(tc1);
      deref_cell(tc2);
      return ret;

    case EXPR_MULT:
      x = cell2int(lexic, st->link[0]);
      y = cell2int(lexic, st->link[1]);
      return int2cell(x * y);

    case EXPR_DIV:
      x = cell2int(lexic, st->link[0]);
      y = cell2int(lexic, st->link[1]);
      if( y != 0 )
       return bool2cell(x / y);
      else
       return bool2cell(0);
       
    case EXPR_MODULO:
      x = cell2int(lexic, st->link[0]);
      y = cell2int(lexic, st->link[1]);
      if( y != 0)
       return int2cell(x % y);
      else
       return int2cell(0);

    case EXPR_BIT_AND:
      x = cell2int(lexic, st->link[0]);
      y = cell2int(lexic, st->link[1]);
      return bool2cell(x & y);

    case EXPR_BIT_OR:
      x = cell2int(lexic, st->link[0]);
      y = cell2int(lexic, st->link[1]);
      return bool2cell(x | y);

    case COMP_MATCH:
      s1 = cell2str(lexic, st->link[0]);
      s2 = cell2str(lexic, st->link[1]);
      flag = strstr(s2, s1) != NULL;
      free(s1); free(s2);
      return bool2cell(flag);

    case COMP_LT:
      return bool2cell(cell_cmp(lexic, st->link[0], st->link[1]) < 0);

    case COMP_LE:
      return bool2cell(cell_cmp(lexic, st->link[0], st->link[1]) <= 0);

    case COMP_EQ:
      return bool2cell(cell_cmp(lexic, st->link[0], st->link[1]) == 0);

    case COMP_NE:
      return bool2cell(cell_cmp(lexic, st->link[0], st->link[1]) != 0);

    case COMP_GT:
      return bool2cell(cell_cmp(lexic, st->link[0], st->link[1]) > 0);

    case COMP_GE:
      return bool2cell(cell_cmp(lexic, st->link[0], st->link[1]) >= 0);

    case CONST_INT:
    case CONST_STR:
    case CONST_DATA:
      ref_cell(st);	/* nasl_exec returns a cell that should be deref-ed */
      return st;

    case REF_VAR:
    case REF_ARRAY:
      ret = nasl_read_var_ref(lexic, st);
      return ret;

    default:
      fprintf(stderr, "nasl_exec: unhandled node type %d\n", st->type);
      abort();
      return NULL;
    }

 error:
  deref_cell(ret);
  deref_cell(ret2);
  return NULL;
}



int
execute_nasl_script(struct arglist * script_infos, const char* name, int mode)
{
  naslctxt	ctx;
  int		err = 0;
  tree_cell	*ret;
  lex_ctxt	*lexic;
  char 	 	old_dir[MAXPATHLEN+1];
  char		*newdir;
  char		*old;
  tree_cell	description;
  struct arglist*	prefs = arg_get_value(script_infos, "preferences");
  char		*str;
  int to;
  
  nessus_init_random();

  old_dir[sizeof(old_dir) - 1] = '\0';
  getcwd(old_dir, sizeof(old_dir) - 1);
#if NASL_DEBUG > 2
  nasl_trace_fp = stderr;
#endif
 if((old = arg_get_value(script_infos, "script_name")) == NULL)
   arg_add_value(script_infos, "script_name", ARG_STRING, strlen(name), estrdup(name));
 else
   { 
   efree(&old);
   arg_set_value(script_infos, "script_name", strlen(name), estrdup(name));
  }
 
 newdir = strrchr(name, '/');
 if(newdir != NULL)
 {
	 char dir[MAXPATHLEN+1];
	 char *s;
	 dir[sizeof(dir) - 1] = '\0';
	 strncpy(dir, name, sizeof(dir) - 1);
	 s = strrchr(dir, '/');
	 s[0] = '\0';
	 chdir(dir);
 }

 if (init_nasl_ctx(&ctx, name) < 0)
 {
    chdir(old_dir);
    return -1;
 }

 if (naslparse(&ctx))
   {
     fprintf(stderr, "\nParse error at or near line %d\n", ctx.line_nb);
     nasl_clean_ctx(&ctx);
     chdir(old_dir);
     return -1;
   }
#if NASL_DEBUG > 4
 nasl_dump_tree(ctx.tree);
#endif
 lexic = init_empty_lex_ctxt();
 lexic->script_infos = script_infos;
 
 str = arg_get_value(prefs, "checks_read_timeout");
 if( str != NULL )
 	to = atoi(str);
 else
 	to = 5;
	
 if(to <= 0)to = 5;
 
 lexic->recv_timeout = to;
 init_nasl_library(lexic);

 
 bzero(&description, sizeof(description));
 description.type = CONST_INT;
 description.x.i_val = mode;
 add_named_var_to_ctxt(lexic, "description", &description);

 truc = (lex_ctxt*)ctx.tree;
 if ((ret = nasl_exec(lexic, ctx.tree)) == NULL)
   err = -1;
 else
   deref_cell(ret);
#if NASL_DEBUG > 3
 nasl_dump_tree(ctx.tree);
#endif
 nasl_clean_ctx(&ctx);
 free_lex_ctxt(lexic);
 chdir(old_dir);
 return err;
}

