/* Nessus Attack Scripting Language version 2
 *
 * Copyright (C) 2002 Michel Arboi
 *
 * En quelques mots : ceci n'est pas un logiciel libre.
 *
 * Ce logiciel "NASL2" n'est pas termin. Il reste ma proprit jusqu' ce 
 * que je dcide de changer la licence.
 * En attendant, vous avez le droit de linker avec le code sous GPL "Nessus", 
 * provenant de http://www.nessus.org,  fin de tests ou pour votre propre 
 * usage ducatif ou priv.
 * Vous n'avez pas le droit de revendre sous quelque forme que ce soit le
 * logiciel NASL2 ou ses rsultats ; cette restriction s'applique  tout 
 * logiciel link avec NASL2, ou incluant une partie de code provenant de
 * NASL2.
 *
 * Ce programme est distribu en esprant qu'il sera utile, mais SANS AUCUNE
 * GARANTIE.
 * Si vous n'acceptez pas ce point, veuillez me contacter pour acheter une
 * licence commerciale.
 *
 * [quick & dirty english translation. In case of conflict, French version
 *  prevails]
 *
 * In a few words: this is *not* free software.
 *
 * This "NASL2" software is unfinished. It is proprietary software until 
 * I decide to switch to another licence.
 * Meanwhile, you may link it with the GPL-ed "Nessus" software, from 
 * http://www.nessus.org, and use it for testing or for your own edutionnal 
 * or private use.
 * You are not allow to sell either this NASL2 software or its outputs;
 * the same restriction applies to any software linked with NASL2, or that
 * includes a part of NASL2 code.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * If you do not accept this last point, please contact me and buy a commercial
 * licence.
 */

#include <includes.h>


#include "nasl_tree.h"
#include "nasl_global_ctxt.h"
#include "nasl_func.h"
#include "nasl_var.h"
#include "nasl_lex_ctxt.h"
#include "exec.h"

#ifndef NASL_DEBUG
#define NASL_DEBUG 0
#endif

static int
hash_int(int x)
{
  return (unsigned) x % VAR_NAME_HASH;
}

static int
hash_str(const char* s)
{
  const char	*p;
  int		h = 0;

  if (s == NULL)
    return 0;
  for (p = s; *p != '\0'; p ++)
    h = (h * (VAR_NAME_HASH-1) + *p) % VAR_NAME_HASH;
  return h;
}


static nasl_var*
get_var_by_num(nasl_var** vars, int num)
{
  int		h = hash_int(num);
  nasl_var	*v;

  for (v = vars[h]; v != NULL; v = v->next_var)
    if (v->var_name == NULL && num == v->var_num)
      return v;

  v = emalloc(sizeof(nasl_var));
  v->var_num = num;
  v->type = VAR2_UNDEF;
  v->next_var = vars[h];

  vars[h] = v;
  return v;  
}

static nasl_var*
get_var_by_name(nasl_var** vars, const char* s)
{
  int		h = hash_str(s);
  nasl_var	*v;
  
  for (v = vars[h]; v != NULL; v = v->next_var)
    if (v->var_name != NULL && strcmp(s, v->var_name) == 0)
      return v;

  v = emalloc(sizeof(nasl_var));
  v->var_name = estrdup(s);
  v->type = VAR2_UNDEF;
  v->next_var = vars[h];

  vars[h] = v;
  return v;  
}



/* This function climbs up in the context list */
static nasl_var*
get_var_ref_by_name(lex_ctxt* ctxt, const char* name, int climb)
{
  nasl_var	*v;
  char		*p;
  int	h = hash_str(name);
  lex_ctxt	*c;
  harglst	*globals, * variables, *types;
  int		vtype;

  if(climb != 0)
  {
  for (c = ctxt; c != NULL; c = c->up_ctxt)
    {
      for (v = c->variables[h]; v != NULL; v = v->next_var)
	if (v->var_name != NULL && strcmp(name, v->var_name) == 0)
	  return v;
    }
  }
 else
 {
   for (v = ctxt->variables[h]; v != NULL; v = v->next_var)
	if (v->var_name != NULL && strcmp(name, v->var_name) == 0)
	  return v;
 }

  v = emalloc(sizeof(nasl_var));
  v->var_name = estrdup(name);
  v->type = VAR2_UNDEF;
  v->next_var = ctxt->variables[h];
  ctxt->variables[h] = v;

  return v;  
}


static nasl_var*
get_var_ref_by_num(lex_ctxt* ctxt, int num)
{
  nasl_var	*v;
  int	h = hash_int(num);

  /* We do not climb the contexts, this does not make sense */
  for (v = ctxt->variables[h]; v != NULL; v = v->next_var)
    if (v->var_name == NULL && v->var_num == num)
      return v;

  v = emalloc(sizeof(nasl_var));
  v->var_num = num;
  v->type = VAR2_UNDEF;
  v->next_var = ctxt->variables[h];

  ctxt->variables[h] = v;
  return v;  
}

static tree_cell*
var2cell(nasl_var* v)
{
  tree_cell	*tc = alloc_tree_cell(0, NULL);

  tc->type = REF_VAR;
  tc->x.ref_val = v;		/* No need to free this later! */
  return tc;
}

tree_cell*
get_variable_by_name(lex_ctxt* ctxt, const char* name)
{
  nasl_var	* v = get_var_ref_by_name(ctxt, name, 1);
  return var2cell(v);
}

const char*
get_var_name(nasl_var *v)
{
  static char	n[32];

  if (v->var_name != NULL)
    return v->var_name;
  sprintf(n, "#%d", v->var_num);
  return n;
}

tree_cell*
get_array_elem(lex_ctxt* ctxt, const char* name, tree_cell* idx)
{
  nasl_var	*a = get_var_ref_by_name(ctxt, name, 1);
  nasl_var	*e;
  char		txt[32], *idxname;
  int		h;
  tree_cell	*tc;
  lex_ctxt	lexic2;


  if (idx == NULL)
    {
#if NASL_DEBUG > 0
      fprintf(stderr, "get_array_elem: NULL index\n");
#endif
      return NULL;
    }

  switch (a->type)
    {
    case VAR2_UNDEF:
      /* We define the array here */
      a->type = VAR2_ARRAY;
    case VAR2_ARRAY:
      if (a->x.a_val == NULL)
	a->x.a_val = emalloc(VAR_NAME_HASH * sizeof(nasl_var*));

      switch(idx->type)
	{
	case CONST_INT:
	  e = get_var_by_num(a->x.a_val, idx->x.i_val);
	  return var2cell(e);
	  
	case CONST_STR:
	case CONST_DATA:
	  e = get_var_by_name(a->x.a_val, idx->x.str_val);
	  return var2cell(e);
	  
	default:
	  fprintf(stderr, "get_array_elem: unhandled index type 0x%x\n",
		  idx->type);
	  abort();
	  return NULL;
	}
      /*NOTREACHED*/
      break;

    case VAR2_INT:
      fprintf(stderr, "get_array_elem: variable %s is an integer\n",
	      get_var_name(a));
      return NULL;

    case VAR2_STRING:
    case VAR2_DATA:
      if (idx->type == CONST_INT)
	{
	  int	l = a->size;

	  if (idx->x.i_val >= l)
	    {
	      fprintf(stderr, "get_array_elem: requesting character after end of string %s (%d >= %d)\n", get_var_name(a), idx->x.i_val, l);
	      tc = alloc_expr_cell(idx->line_nb, CONST_DATA /*CONST_STR*/,
				   NULL, NULL);
	      tc->x.str_val = estrdup("");
	      tc->size = 0;
	      return tc;
	    }
	  else
	    {
	      tc = alloc_expr_cell(idx->line_nb, CONST_DATA /*CONST_STR*/,
				   NULL, NULL);
	      tc->x.str_val = emalloc(2);
	      tc->x.str_val[0] = a->x.str_val[idx->x.i_val];
	      tc->x.str_val[1] = '\0';
	      tc->size = 1;
	      return tc;
	    }
	}
      else
	{
	  fprintf(stderr, "get_array_elem: Cannot use a non integer index  (type 0x%x) in string\n", idx->type);
	  return NULL;
	}
      /*NOTREACHED*/
      break;

    default:
      fprintf(stderr, "Severe bug: unknown variable type 0x%x %s\n",
	      a->type, get_line_nb(idx));
      return NULL;
    }
  /*NOTREACHED*/
  return NULL;
}


void
free_var_chain(nasl_var* v)
{
  int	i;

  if (v == NULL)
    return;
  free_var_chain(v->next_var);
  efree(&v->var_name);
  switch(v->type)
    {
    case VAR2_STRING:
    case VAR2_DATA:
      efree(&v->x.str_val);
      break;
    case VAR2_ARRAY:
      for (i = 0; i < VAR_NAME_HASH; i ++)
	free_var_chain(v->x.a_val[i]);
      efree(&v->x.a_val);
      break;
    }
  efree(&v);
}

void
clear_var(nasl_var* v)
{
  int	i;

  if (v == NULL)
    return;

  switch (v->type)
    {
    case VAR2_INT:
      v->x.i_val = 0;
      break;
    case VAR2_STRING:
    case VAR2_DATA:
      efree(&v->x.str_val);
      break;
    case VAR2_ARRAY:
      for (i = 0; i < VAR_NAME_HASH; i ++)
	free_var_chain(v->x.a_val[i]);
      efree(&v->x.a_val);
      break;
    }
  v->type = VAR2_UNDEF;
  return;
}

extern FILE*	nasl_trace_fp;

tree_cell*
nasl_affect(tree_cell* lval, tree_cell* rval)
{
  nasl_var	*v1 = NULL, *v2 = NULL;
  int		t1, t2;


  if (lval->type != REF_VAR)
    {
      fprintf(stderr, "nasl_affect: cannot affect to non variable %s\n",
	      nasl_type_name(lval->type));
      return NULL;
    }

  v1 = lval->x.ref_val;
  t1 = v1->type;

  if (rval == NULL || rval == FAKE_CELL)
    {
#if NASL_DEBUG > 0
      fprintf(stderr, "nasl_affect: cannot affect NULL or FAKE cell - undefining variable '%s' %s\n", get_var_name(v1), get_line_nb(lval));
#endif
      clear_var(v1);
      if (nasl_trace_fp != NULL)
	fprintf(nasl_trace_fp, "NASL> %s <- undef\n", get_var_name(v1));

      return NULL;
    }

    switch (rval->type)
      {
      case CONST_INT:
	t2 = VAR2_INT;
	break;
      case CONST_STR:
	t2 = VAR2_STRING;
	break;
      case CONST_DATA:
	t2 = VAR2_DATA;
	break;

      case REF_VAR:
	v2 = rval->x.ref_val;
	t2 = v2->type;
	break;

      default:
	fprintf(stderr, "Cannot affect rvalue 0x%x to variable\n", rval->type);
	return NULL;
      }

  clear_var(v1);

  if (t1 != VAR2_UNDEF && t2 == VAR2_UNDEF)
    {
#if NASL_DEBUG > 0
      fprintf(stderr, 
	      "Warning: Undefining defined variable %s\n",get_var_name(v1));
#endif
    }
  else if (t1 == VAR2_ARRAY && t2 != VAR2_ARRAY)
    {
#if NASL_DEBUG > 1
      fprintf(stderr, 
	      "Warning: affecting non array (0x%x) to array variable %s\n",
	      t2, get_var_name(v1));
#endif
    }
  else if ((t1 == VAR2_INT || t1 == VAR2_STRING || t1 == VAR2_DATA) &&
	   t2 == VAR2_ARRAY)
    {
#if NASL_DEBUG > 1
      fprintf(stderr, "Warning: affecting array to atomic variable (0x%x)\n",
	      get_var_name(v1), t2);
#endif
    }

  v1->type = t2;
  if (rval->type != REF_VAR)
    switch (t2)
      {
      case VAR2_INT:
	v1->x.i_val = rval->x.i_val;
	break;
      case VAR2_STRING:
      case VAR2_DATA:
	v1->x.str_val = emalloc(rval->size);
	bcopy(rval->x.str_val, v1->x.str_val, rval->size);
	v1->size = rval->size;
	break;
      }
  else				/* REF_VAR */
    switch(t2)
      {
      case VAR2_INT:
	v1->x.i_val = v2->x.i_val;
	break;
      case VAR2_STRING:
      case VAR2_DATA:
	v1->x.str_val = emalloc(v2->size);
	bcopy(v2->x.str_val, v1->x.str_val, v2->size);
	v1->size = v2->size;
	break;
      case VAR2_ARRAY:
	/* TBD */
	fprintf(stderr, "nasl_affect: Arrays not handled yet\n");
	abort();
	break;
      }

  if (nasl_trace_fp != NULL)
    switch(t2)
      {
      case VAR2_INT:
	fprintf(nasl_trace_fp, "NASL> %s <- %d\n", get_var_name(v1), v1->x.i_val);
	break;
      case VAR2_STRING:
      case VAR2_DATA:
	fprintf(nasl_trace_fp, "NASL> %s <- \"%s\"\n", get_var_name(v1), v1->x.str_val);
	break;
      default:
	fprintf(nasl_trace_fp, "NASL> %s <- ????\n", get_var_name(v1));
	break;
      }

  return FAKE_CELL;
}


static nasl_var*
create_var(const char* name, tree_cell* val)
{
  nasl_var	*v = emalloc(sizeof(nasl_var)), *v2;
  int		t2;

  if (name != NULL)
    v->var_name = estrdup(name);

  if (val == NULL || val == FAKE_CELL)
    {
#if NASL_DEBUG > 0
      fprintf(stderr, "create_var: should not affect NULL or FAKE cell to variable\n");
#endif
      v->type = VAR2_UNDEF;
      return v;
    }

  /* Duplicated code - should be cleaned! */
  switch (val->type)
    {
    case CONST_INT:
      v->type = VAR2_INT;
      v->x.i_val = val->x.i_val;
      break;
    case CONST_STR:
      v->type = VAR2_STRING;
      /* Fix bad string length */
      if (val->size <= 0)
	v->size = strlen(val->x.str_val);
      else
	v->size = val->size;
      v->x.str_val = emalloc(val->size + 1);
      memcpy(v->x.str_val, val->x.str_val, val->size);
      break;
    case CONST_DATA:
      v->type = VAR2_DATA;
      v->size = val->size;
      v->x.str_val = emalloc(val->size);
      memcpy(v->x.str_val, val->x.str_val, val->size);
      break;
    case REF_VAR:
      v2 = val->x.ref_val;
      v->type = v2->type;
      v->x = v2->x;
      break;

    default:
      fprintf(stderr, "create_var: cannot affect rvalue 0x%x to variable\n",
	      val->type);
#if 1
      v->type = VAR2_UNDEF;
#else
      efree(&v);
      return NULL;
#endif
    }
  return v;
}

tree_cell* decl_local_variables(lex_ctxt* lexic, tree_cell* vars)
{
  tree_cell	*t;

  for (t = vars; t != NULL; t = t->link[0])
    if (t->x.str_val == NULL)
      fprintf(stderr, "decl_local_variables: null name!\n");
    else
      add_named_var_to_ctxt(lexic, t->x.str_val, NULL);
  return FAKE_CELL;
}

tree_cell*
decl_global_variables(lex_ctxt* lexic, tree_cell* vars)
{
  lex_ctxt	*c = lexic;

  while (c->up_ctxt != NULL)
    c = c->up_ctxt;
  return decl_local_variables(c, vars);
}

nasl_var*
add_numbered_var_to_ctxt(lex_ctxt* lexic, int num, tree_cell* val)
{
  int		h = hash_int(num);
  nasl_var	*v;
  
  for (v = lexic->variables[h]; v != NULL; v = v->next_var)
    if (v->var_name == NULL && num == v->var_num)
      {
	fprintf(stderr, "Cannot add existing variable %d\n", num);
	return NULL;
      }
  v = create_var(NULL, val);
  v->var_num = num;
  v->next_var = lexic->variables[h];
  lexic->variables[h] = v;
  return v;
}

nasl_var*
add_named_var_to_ctxt(lex_ctxt* lexic, const char* name, tree_cell* val)
{
  int		h = hash_str(name);
  nasl_var	*v;
  
  /* Duplicated code - again */
  for (v = lexic->variables[h]; v != NULL; v = v->next_var)
    if (v->var_name != NULL && strcmp(name, v->var_name) == 0)
      {
	fprintf(stderr, "Cannot add existing variable %s\n", name);
	return NULL;
      }
  v = create_var(name, val);
  if (v == NULL)
    return NULL;
  v->next_var = lexic->variables[h];
  lexic->variables[h] = v;
  return v;
}



tree_cell*
nasl_read_var_ref(lex_ctxt* lexic, tree_cell* tc)
{
  tree_cell	*ret;
  nasl_var	*v;

  if (tc == NULL || tc == FAKE_CELL)
    {
      fprintf(stderr, "nasl_read_var_ref: cannot read NULL or FAKE cell\n");
      return NULL;
    }
  if (tc->type != REF_VAR)
    {
      fprintf(stderr,
	      "nasl_read_var_ref: argument (type=%d) is not REF_VAR %s\n",
	      tc->type, get_line_nb(tc));
      return NULL;
    }

  v = tc->x.ref_val;
  ret = alloc_tree_cell(tc->line_nb, NULL);

  if (nasl_trace_fp != NULL)
    fprintf(nasl_trace_fp, "NASL> %s -> ", get_var_name(v));

  switch (v->type)
    {
    case VAR2_INT:
      ret->type = CONST_INT;
      ret->x.i_val = v->x.i_val;
      if (nasl_trace_fp != NULL)
	fprintf(nasl_trace_fp, "%d\n", ret->x.i_val);
      return ret;

    case VAR2_STRING:
      ret->type = CONST_STR;
      ret->x.str_val = estrdup(v->x.str_val);
      /* Fix bad string length */
      if (v->size <= 0)
	ret->size = strlen(v->x.str_val);
      else
	ret->size = v->size;
      if (nasl_trace_fp != NULL)
	fprintf(nasl_trace_fp, "\"%s\"\n", ret->x.str_val);
      return ret;

    case VAR2_DATA:
      ret->type = CONST_DATA;
      ret->x.str_val = emalloc(v->size);
      memcpy(ret->x.str_val, v->x.str_val, v->size);
      ret->size = v->size;
      if (nasl_trace_fp != NULL)
	fprintf(nasl_trace_fp, "\"%s\"\n", ret->x.str_val);
      return ret;

    case VAR2_ARRAY:
      fprintf(stderr, "nasl_read_var_ref: cannot read whole array '%s'\n",
	      get_var_name(v));
      break;

    case VAR2_UNDEF:
#if NASL_DEBUG > 0
      fprintf(stderr, "nasl_read_var_ref: variable '%s' is undefined %s\n",
	      get_var_name(v), get_line_nb(tc));
#endif
      break;

    default:
      fprintf(stderr, "nasl_read_var_ref: unhandled variable type %d\n",
	      v->type);
      break;
    }
  deref_cell(ret);
  if (nasl_trace_fp != NULL)
    fprintf(nasl_trace_fp, "????\n");
  return NULL;
}


tree_cell*
nasl_incr_variable(lex_ctxt* lexic, tree_cell* tc, int pre, int val)
{
  nasl_var	*v;
  int		old_val = 0, new_val;
  tree_cell	*retc;

  if (tc->type != REF_VAR)
    {
      fprintf(stderr,
	      "nasl_incr_variable: argument (type=%d) is not REF_VAR %s\n",
	      tc->type, get_line_nb(tc));
      return NULL;
    }

  v = tc->x.ref_val;

  switch (v->type)
    {
    case VAR2_INT:
      old_val = v->x.i_val;
      break;
    case VAR2_STRING:
    case VAR2_DATA:
#if NASL_DEBUG > 0
      fprintf(stderr, "nasl_incr_variable: variable %s is a STRING %s - converting to integer\n", get_var_name(v), get_line_nb(tc));
#endif
      old_val = v->x.str_val == NULL ? 0 : atoi(v->x.str_val);
      break;
    case VAR2_UNDEF:
#if NASL_DEBUG > 0
      fprintf(stderr, "nasl_incr_variable: variable %s is undefined %s\n",
	      get_var_name(v), get_line_nb(tc));
#endif
      old_val = 0;
      break;

    default:
      fprintf(stderr, "nasl_incr_variable: variable %s has bad type %d %s\n",
	      get_var_name(v), get_line_nb(tc));
      return NULL;
    }
  new_val = old_val + val;

  clear_var(v);
  v->type = VAR2_INT;
  v->x.i_val = new_val;

  retc = alloc_tree_cell(0, NULL);
  retc->type = CONST_INT;
  retc->x.i_val = pre ? new_val : old_val;
  
  return retc;  
}


static int	var2int(nasl_var* v, int defval)
{
  if (v == NULL)
    return defval;

  switch (v->type)
    {
    case VAR2_INT:
      return v->x.i_val;

    case VAR2_STRING:
    case VAR2_DATA:
      return atoi(v->x.str_val);

    case VAR2_UNDEF:
#if NASL_DEBUG > 0
      fprintf(stderr, "var2int: variable '%s' is undefined!\n", 
	      get_var_name(v));
#endif
    case VAR2_ARRAY:
    default:
      return defval;
    }
}

const char*	var2str(nasl_var* v)
{
  static char	s1[16];

  if (v == NULL)
    return NULL;

  switch (v->type)
    {
    case VAR2_INT:
      sprintf(s1, "%d", v->x.i_val);
      return s1;		/* dangerous if called twice in a row */

    case VAR2_STRING:
    case VAR2_DATA:
      return v->x.str_val;

    case VAR2_UNDEF:
#if NASL_DEBUG > 0
      fprintf(stderr, "var2str: variable '%s' is undefined!\n", 
	      get_var_name(v));
#endif
    case VAR2_ARRAY:
    default:
      return NULL;
    }
}

int
get_int_var_by_num(lex_ctxt* lexic, int num, int defval)
{
  nasl_var	*v = get_var_ref_by_num(lexic, num);
  return  var2int(v, defval);
}

int
get_int_var_by_name(lex_ctxt* lexic, const char* name, int defval)
{
  nasl_var	*v = get_var_ref_by_name(lexic, name, 1);
  return  var2int(v, defval);
}

int 
get_int_local_var_by_name(lex_ctxt * lexic, const char * name, int defval)
{
 nasl_var 	*v = get_var_ref_by_name(lexic, name, 0);
 return var2int(v, defval);
}


char*
get_str_var_by_num(lex_ctxt* lexic, int num)
{
  nasl_var	*v = get_var_ref_by_num(lexic, num);
  return  (char*)var2str(v);
}

char*
get_str_var_by_name(lex_ctxt* lexic, const char* name)
{
  nasl_var	*v = get_var_ref_by_name(lexic, name, 1);
  return  (char*)var2str(v);
}

char *
get_str_local_var_by_name(lex_ctxt * lexic, const char * name)
{
  nasl_var	*v = get_var_ref_by_name(lexic, name, 0);
  return  (char*)var2str(v);
}




int
get_var_size_by_name(lex_ctxt* lexic, const char* name)
{
  nasl_var	*v = get_var_ref_by_name(lexic, name, 1);
  return v == NULL ? 0 : v->size;
}

int
get_local_var_size_by_name(lex_ctxt * lexic, const char * name)
{
 nasl_var 	*v = get_var_ref_by_name(lexic, name, 0);
 return v == NULL ? 0 : v->size;
}

int
get_var_size_by_num(lex_ctxt* lexic, int num)
{
  nasl_var	*v = get_var_ref_by_num(lexic, num);
  return v == NULL ? 0 : v->size;
}


int
get_var_type_by_num(lex_ctxt* lexic, int num)
{
  nasl_var	*v = get_var_ref_by_num(lexic, num);
  return v == NULL ? 0 : v->type;
}

int
get_var_type_by_name(lex_ctxt* lexic, const char * name)
{
  nasl_var	*v = get_var_ref_by_name(lexic, name, 1);
  return v == NULL ? 0 : v->type;
}

int get_local_var_type_by_name(lex_ctxt * lexic, const char * name)
{
 nasl_var	*v = get_var_ref_by_name(lexic, name, 0);
 return v == NULL ? 0 : v->type;
}
